@@ -10,6 +10,7 @@ use clap::Parser;
 use log::{debug, warn};
 use std::cmp::Ordering;
 use std::collections::BTreeMap;
+use std::fmt::Write;
 use std::fs;
 use std::io::Read;
 use std::path::PathBuf;
@@ -64,7 +65,7 @@ fn define_command_helper(
 
     if let Some(jsonfile) = jsonfile {
         let _ = std::fs::File::open(&jsonfile)
-            .with_context(|| format!("Unable to read file {:?}", jsonfile));
+            .with_context(|| format!("Unable to read file {:?}", jsonfile))?;
 
         if mdev_type.is_some() {
             return Err(anyhow!(
@@ -81,7 +82,7 @@ fn define_command_helper(
             return Err(anyhow!(
                 "Cowardly refusing to overwrite existing config for {}/{}",
                 parent,
-                uuid.to_hyphenated().to_string()
+                uuid.hyphenated().to_string()
             ));
         }
 
@@ -115,7 +116,7 @@ fn define_command_helper(
         if dev.is_defined() {
             return Err(anyhow!(
                 "Device {} on {} already defined",
-                dev.uuid.to_hyphenated().to_string(),
+                dev.uuid.hyphenated().to_string(),
                 dev.parent()?
             ));
         }
@@ -132,6 +133,7 @@ fn define_command(
     parent: Option<String>,
     mdev_type: Option<String>,
     jsonfile: Option<PathBuf>,
+    force: bool,
 ) -> Result<()> {
     debug!("Defining mdev {:?}", uuid);
 
@@ -143,32 +145,73 @@ fn define_command(
         and populate the attrs field correspondingly before the device is defined in the system.
         The device config file will contain the same attributes that were used to start this device。
     */
-    if dev.active {
-        let attrs = Callout::get_attributes(&mut dev)?;
-        dev.add_attributes(&attrs)?;
-    }
-    Callout::invoke(&mut dev, Action::Define, |dev| dev.define()).map(|_| {
+    let mut c = callout(&mut dev);
+    c.invoke(Action::Define, force, |c| {
+        if c.dev.active {
+            let attrs = c.get_attributes()?;
+            c.dev.add_attributes(&attrs)?;
+        }
+        c.dev.define()
+    })
+    .map(|_| {
         if uuid.is_none() {
-            println!("{}", dev.uuid.to_hyphenated());
+            println!("{}", dev.uuid.hyphenated());
         }
     })
 }
 
 /// Implementation of the `mdevctl undefine` command
-fn undefine_command(env: &dyn Environment, uuid: Uuid, parent: Option<String>) -> Result<()> {
+fn undefine_command(
+    env: &dyn Environment,
+    uuid: Uuid,
+    parent: Option<String>,
+    force: bool,
+) -> Result<()> {
     debug!("Undefining mdev {:?}", uuid);
+    let mut failed = false;
     let devs = defined_devices(env, Some(&uuid), parent.as_ref())?;
     if devs.is_empty() {
         return Err(anyhow!("No devices match the specified uuid"));
     }
     for (_, mut children) in devs {
         for child in children.iter_mut() {
-            let _ = Callout::invoke(child, Action::Undefine, |dev| dev.undefine());
+            let mut c = callout(child);
+            if let Err(e) = c.invoke(Action::Undefine, force, |c| c.dev.undefine()) {
+                failed = true;
+                for x in e.chain() {
+                    warn!(
+                        "Undefine of {} on parent {} failed with error: {}",
+                        c.dev.uuid,
+                        c.dev.parent().unwrap().to_string(),
+                        x
+                    );
+                }
+            }
         }
     }
+    if failed {
+        return Err(anyhow!("Undefine failed"));
+    }
     Ok(())
 }
 
+fn dev_from_jsonfile(
+    env: &dyn Environment,
+    uuid: Uuid,
+    parent: String,
+    jsonfile: PathBuf,
+) -> Result<MDev, anyhow::Error> {
+    let _ = std::fs::File::open(&jsonfile)
+        .with_context(|| format!("Unable to read file {:?}", jsonfile))?;
+    let filecontents = fs::read_to_string(&jsonfile)
+        .with_context(|| format!("Unable to read jsonfile {:?}", jsonfile))?;
+    let jsonval = serde_json::from_str(&filecontents)?;
+
+    let mut d = MDev::new(env, uuid);
+    d.load_from_json(parent, &jsonval)?;
+    Ok(d)
+}
+
 /// Implementation of the `mdevctl modify` command
 #[allow(clippy::too_many_arguments)]
 fn modify_command(
@@ -182,31 +225,91 @@ fn modify_command(
     value: Option<String>,
     auto: bool,
     manual: bool,
+    live: bool,
+    defined: bool,
+    jsonfile: Option<PathBuf>,
+    force: bool,
 ) -> Result<()> {
-    let mut dev = get_defined_device(env, uuid, parent.as_ref())?;
-    if mdev_type.is_some() {
-        dev.mdev_type = mdev_type;
-    }
+    debug!("Modifying mdev {:?}", uuid);
+    if live {
+        if mdev_type.is_some() {
+            return Err(anyhow!("'type' cannot be changed on active mdev"));
+        }
+        if auto {
+            return Err(anyhow!("'auto' cannot be changed on active mdev"));
+        }
+        if manual {
+            return Err(anyhow!("'manual' cannot be changed on active mdev"));
+        }
+        let mut act_dev = get_active_device(env, uuid, parent.as_ref())?;
+        if let Some(f) = jsonfile {
+            let act_parent = act_dev
+                .parent
+                .clone()
+                .ok_or_else(|| anyhow!("Parent device required to modify device via json file"))?;
+            let json_dev = dev_from_jsonfile(env, uuid, act_parent, f)?;
+            if json_dev.mdev_type != act_dev.mdev_type {
+                return Err(anyhow!("'type' cannot be changed on active mdev"));
+            }
+            if json_dev.parent != act_dev.parent {
+                return Err(anyhow!("'parent' cannot be changed on active mdev"));
+            }
+            act_dev = json_dev;
+        } else {
+            return Err(anyhow!("'live' option must be used with 'jsonfile' option"));
+        }
 
-    if auto {
-        dev.autostart = true;
-    } else if manual {
-        dev.autostart = false;
-    }
+        if defined {
+            // live and stored modify - defined dev config exists and types match
+            let def_dev = get_defined_device(env, uuid, act_dev.parent.as_ref())?;
+            if def_dev.mdev_type != act_dev.mdev_type {
+                return Err(anyhow!("'type' of active and defined mdev does not match"));
+            }
 
-    match addattr {
-        Some(attr) => match value {
-            None => return Err(anyhow!("No attribute value provided")),
-            Some(v) => dev.add_attribute(attr, v, index)?,
-        },
-        None => {
-            if delattr {
-                dev.delete_attribute(index)?;
+            let mut c = callout(&mut act_dev);
+            debug!("mdev device used for live update '{:?}'", c.dev);
+            return c
+                .invoke_modify_live()
+                .and_then(|_| c.invoke(Action::Modify, force, |c| c.dev.write_config()));
+        }
+        // live modify only
+        callout(&mut act_dev).invoke_modify_live()
+    } else {
+        let mut dev: MDev;
+        // stored configuration modify
+        if let Some(f) = jsonfile {
+            let parent = parent
+                .ok_or_else(|| anyhow!("Parent device required to modify device via json file"))?;
+            dev = dev_from_jsonfile(env, uuid, parent, f)?;
+        } else {
+            dev = get_defined_device(env, uuid, parent.as_ref())?;
+            if mdev_type.is_some() {
+                dev.mdev_type = mdev_type;
+            }
+            if auto && manual {
+                return Err(anyhow!("'auto' and 'manual' are mutually exclusive"));
+            }
+            if auto {
+                dev.autostart = true;
+            } else if manual {
+                dev.autostart = false;
             }
         }
-    }
 
-    Callout::invoke(&mut dev, Action::Modify, |dev| dev.write_config())
+        let index = index.map(|n| n as usize);
+        match addattr {
+            Some(attr) => match value {
+                None => return Err(anyhow!("No attribute value provided")),
+                Some(v) => dev.add_attribute(attr, v, index)?,
+            },
+            None => {
+                if delattr {
+                    dev.delete_attribute(index)?;
+                }
+            }
+        }
+        callout(&mut dev).invoke(Action::Modify, force, |c| c.dev.write_config())
+    }
 }
 
 /// convert 'start' command arguments into a MDev struct
@@ -247,7 +350,7 @@ fn start_command_helper(
                     Ordering::Greater => {
                         return Err(anyhow!(
                             "Multiple definitions found for device {}. Please specify a parent.",
-                            uuid.unwrap().to_hyphenated().to_string()
+                            uuid.unwrap().hyphenated().to_string()
                         ));
                     }
                     Ordering::Equal => {
@@ -260,7 +363,7 @@ fn start_command_helper(
                             if mdev_type.is_some() && mdev_type != d.mdev_type {
                                 return Err(anyhow!(
                                     "Device {} already exists on parent {} with type {}",
-                                    d.uuid.to_hyphenated().to_string(),
+                                    d.uuid.hyphenated().to_string(),
                                     d.parent().unwrap(),
                                     d.mdev_type.as_ref().unwrap()
                                 ));
@@ -300,23 +403,25 @@ fn start_command(
     parent: Option<String>,
     mdev_type: Option<String>,
     jsonfile: Option<PathBuf>,
+    force: bool,
 ) -> Result<()> {
     let mut dev = start_command_helper(env, uuid, parent, mdev_type, jsonfile)?;
 
-    Callout::invoke(&mut dev, Action::Start, |dev| dev.start()).map(|_| {
+    let mut c = callout(&mut dev);
+    c.invoke(Action::Start, force, |c| c.dev.start()).map(|_| {
         if uuid.is_none() {
-            println!("{}", dev.uuid.to_hyphenated());
+            println!("{}", c.dev.uuid.hyphenated());
         }
     })
 }
 
 /// Implementation of the `mdevctl stop` command
-fn stop_command(env: &dyn Environment, uuid: Uuid) -> Result<()> {
+fn stop_command(env: &dyn Environment, uuid: Uuid, force: bool) -> Result<()> {
     debug!("Stopping '{}'", uuid);
     let mut dev = MDev::new(env, uuid);
     dev.load_from_sysfs()?;
 
-    Callout::invoke(&mut dev, Action::Stop, |dev| dev.stop())
+    callout(&mut dev).invoke(Action::Stop, force, |c| c.dev.stop())
 }
 
 /// convenience function to lookup a defined device by uuid and parent
@@ -327,39 +432,39 @@ fn get_defined_device<'a>(
 ) -> Result<MDev<'a>> {
     let devs = defined_devices(env, Some(&uuid), parent)?;
     if devs.is_empty() {
-        return match parent {
+        match parent {
             None => Err(anyhow!(
                 "Mediated device {} is not defined",
-                uuid.to_hyphenated().to_string()
+                uuid.hyphenated().to_string()
             )),
             Some(p) => Err(anyhow!(
                 "Mediated device {}/{} is not defined",
                 p,
-                uuid.to_hyphenated().to_string()
+                uuid.hyphenated().to_string()
             )),
-        };
+        }
     } else if devs.len() > 1 {
-        return match parent {
+        match parent {
             None => Err(anyhow!(
                 "Multiple definitions found for {}, specify a parent",
-                uuid.to_hyphenated().to_string()
+                uuid.hyphenated().to_string()
             )),
             Some(p) => Err(anyhow!(
                 "Multiple definitions found for {}/{}",
                 p,
-                uuid.to_hyphenated().to_string()
+                uuid.hyphenated().to_string()
             )),
-        };
+        }
     } else {
         let (parent, children) = devs.iter().next().unwrap();
         if children.len() > 1 {
             return Err(anyhow!(
                 "Multiple definitions found for {}/{}",
                 parent,
-                uuid.to_hyphenated().to_string()
+                uuid.hyphenated().to_string()
             ));
         }
-        return Ok(children.get(0).unwrap().clone());
+        Ok(children.first().unwrap().clone())
     }
 }
 
@@ -374,7 +479,7 @@ fn defined_devices<'a>(
         "Looking up defined mdevs: uuid={:?}, parent={:?}",
         uuid, parent
     );
-    for parentpath in env.persist_base().read_dir()?.skip_while(|x| match x {
+    for parentpath in env.config_base().read_dir()?.skip_while(|x| match x {
         Ok(d) => d.path() == env.scripts_base(),
         _ => false,
     }) {
@@ -450,6 +555,114 @@ fn defined_devices<'a>(
     Ok(devices)
 }
 
+/// convenience function to lookup an active device by uuid and parent
+fn get_active_device<'a>(
+    env: &'a dyn Environment,
+    uuid: Uuid,
+    parent: Option<&String>,
+) -> Result<MDev<'a>> {
+    let devs = active_devices(env, Some(&uuid), parent)?;
+    if devs.is_empty() {
+        match parent {
+            None => Err(anyhow!(
+                "Mediated device {} is not active",
+                uuid.hyphenated().to_string()
+            )),
+            Some(p) => Err(anyhow!(
+                "Mediated device {}/{} is not active",
+                p,
+                uuid.hyphenated().to_string()
+            )),
+        }
+    } else if devs.len() > 1 {
+        Err(anyhow!(
+            "Multiple parents found for {}. System error?",
+            uuid.hyphenated().to_string()
+        ))
+    } else {
+        let (parent, children) = devs.iter().next().unwrap();
+        if children.len() > 1 {
+            return Err(anyhow!(
+                "Multiple definitions found for {}/{}",
+                parent,
+                uuid.hyphenated().to_string()
+            ));
+        }
+        Ok(children.first().unwrap().clone())
+    }
+}
+
+/// Get a map of all active devices, optionally filtered by uuid and parent
+fn active_devices<'a>(
+    env: &'a dyn Environment,
+    uuid: Option<&Uuid>,
+    parent: Option<&String>,
+) -> Result<BTreeMap<String, Vec<MDev<'a>>>> {
+    let mut devices: BTreeMap<String, Vec<MDev>> = BTreeMap::new();
+    debug!(
+        "Looking up active mdevs: uuid={:?}, parent={:?}",
+        uuid, parent
+    );
+    if let Ok(dir) = env.mdev_base().read_dir() {
+        for dir_dev in dir {
+            let dir_dev = dir_dev?;
+            let fname = dir_dev.file_name();
+            let basename = fname.to_str().unwrap();
+            debug!("found defined mdev {}", basename);
+            let u = Uuid::parse_str(basename);
+
+            if u.is_err() {
+                warn!("Can't determine uuid for file '{}'", basename);
+                continue;
+            }
+            let u = u.unwrap();
+
+            if uuid.is_some() && uuid != Some(&u) {
+                debug!(
+                    "Ignoring device {} because it doesn't match uuid {}",
+                    u,
+                    uuid.unwrap()
+                );
+                continue;
+            }
+
+            let mut dev = MDev::new(env, u);
+            if dev.load_from_sysfs().is_ok() {
+                if parent.is_some() && (parent != dev.parent.as_ref()) {
+                    debug!(
+                        "Ignoring device {} because it doesn't match parent {}",
+                        dev.uuid,
+                        parent.as_ref().unwrap()
+                    );
+                    continue;
+                }
+
+                // retrieve autostart from persisted mdev if possible
+                let mut per_dev = MDev::new(env, u);
+                per_dev.parent = dev.parent.clone();
+                if per_dev.load_definition().is_ok() {
+                    dev.autostart = per_dev.autostart;
+                }
+
+                // if the device is supported by a callout script that gets attributes, show
+                // those in the output
+                let mut c = callout(&mut dev);
+                if let Ok(attrs) = c.get_attributes() {
+                    let _ = c.dev.add_attributes(&attrs);
+                }
+
+                let devparent = dev.parent()?;
+                if !devices.contains_key(devparent) {
+                    devices.insert(devparent.clone(), Vec::new());
+                };
+
+                devices.get_mut(devparent).unwrap().push(dev);
+            };
+        }
+    }
+    Ok(devices)
+}
+
 /// Implementation of the `mdevctl list` command
 fn list_command(
     env: &dyn Environment,
@@ -473,62 +686,11 @@ fn list_command_helper(
     uuid: Option<Uuid>,
     parent: Option<String>,
 ) -> Result<String> {
-    let mut devices: BTreeMap<String, Vec<MDev>> = BTreeMap::new();
+    let mut devices: BTreeMap<String, Vec<MDev>>;
     if defined {
         devices = defined_devices(env, uuid.as_ref(), parent.as_ref())?;
     } else {
-        debug!("Looking up active mdevs");
-        if let Ok(dir) = env.mdev_base().read_dir() {
-            for dev in dir {
-                let dev = dev?;
-                let fname = dev.file_name();
-                let basename = fname.to_str().unwrap();
-                debug!("found defined mdev {}", basename);
-                let u = Uuid::parse_str(basename);
-
-                if u.is_err() {
-                    warn!("Can't determine uuid for file '{}'", basename);
-                    continue;
-                }
-                let u = u.unwrap();
-
-                if uuid.is_some() && uuid != Some(u) {
-                    debug!(
-                        "Ignoring device {} because it doesn't match uuid {}",
-                        u,
-                        uuid.unwrap()
-                    );
-                    continue;
-                }
-
-                let mut dev = MDev::new(env, u);
-                if dev.load_from_sysfs().is_ok() {
-                    if parent.is_some() && (parent != dev.parent) {
-                        debug!(
-                            "Ignoring device {} because it doesn't match parent {}",
-                            dev.uuid,
-                            parent.as_ref().unwrap()
-                        );
-                        continue;
-                    }
-
-                    let _ = dev.load_definition();
-
-                    if !dev.is_defined() {
-                        if let Ok(attrs) = Callout::get_attributes(&mut dev) {
-                            let _ = dev.add_attributes(&attrs);
-                        }
-                    }
-
-                    let devparent = dev.parent()?;
-                    if !devices.contains_key(devparent) {
-                        devices.insert(devparent.clone(), Vec::new());
-                    };
-
-                    devices.get_mut(devparent).unwrap().push(dev);
-                };
-            }
-        }
+        devices = active_devices(env, uuid.as_ref(), parent.as_ref())?;
     }
 
     // ensure that devices are sorted in a stable order
@@ -668,19 +830,20 @@ fn types_command_helper(
         output.push_str(&jsonstr);
     } else {
         for (parent, children) in types {
-            output.push_str(&format!("{}\n", parent));
+            let _ = writeln!(output, "{}", parent);
             for child in children {
-                output.push_str(&format!("  {}\n", child.typename));
-                output.push_str(&format!(
-                    "    Available instances: {}\n",
+                let _ = writeln!(output, "  {}", child.typename);
+                let _ = writeln!(
+                    output,
+                    "    Available instances: {}",
                     child.available_instances
-                ));
-                output.push_str(&format!("    Device API: {}\n", child.device_api));
+                );
+                let _ = writeln!(output, "    Device API: {}", child.device_api);
                 if !child.name.is_empty() {
-                    output.push_str(&format!("    Name: {}\n", child.name));
+                    let _ = writeln!(output, "    Name: {}", child.name);
                 }
                 if !child.description.is_empty() {
-                    output.push_str(&format!("    Description: {}\n", child.description));
+                    let _ = writeln!(output, "    Description: {}", child.description);
                 }
             }
         }
@@ -709,7 +872,7 @@ fn start_parent_mdevs_command(env: &dyn Environment, parent: String) -> Result<(
         for child in children {
             if child.autostart {
                 debug!("Autostarting {:?}", child.uuid);
-                if let Err(e) = Callout::invoke(child, Action::Start, |child| child.start()) {
+                if let Err(e) = callout(child).invoke(Action::Start, false, |c| c.dev.start()) {
                     for x in e.chain() {
                         warn!("{}", x);
                     }
@@ -754,8 +917,13 @@ fn main() -> Result<()> {
                 parent,
                 mdev_type,
                 jsonfile,
-            } => define_command(&env, uuid, auto, parent, mdev_type, jsonfile),
-            MdevctlCommands::Undefine { uuid, parent } => undefine_command(&env, uuid, parent),
+                force,
+            } => define_command(&env, uuid, auto, parent, mdev_type, jsonfile, force),
+            MdevctlCommands::Undefine {
+                uuid,
+                parent,
+                force,
+            } => undefine_command(&env, uuid, parent, force),
             MdevctlCommands::Modify {
                 uuid,
                 parent,
@@ -766,16 +934,22 @@ fn main() -> Result<()> {
                 value,
                 auto,
                 manual,
+                live,
+                defined,
+                jsonfile,
+                force,
             } => modify_command(
-                &env, uuid, parent, mdev_type, addattr, delattr, index, value, auto, manual,
+                &env, uuid, parent, mdev_type, addattr, delattr, index, value, auto, manual, live,
+                defined, jsonfile, force,
             ),
             MdevctlCommands::Start {
                 uuid,
                 parent,
                 mdev_type,
                 jsonfile,
-            } => start_command(&env, uuid, parent, mdev_type, jsonfile),
-            MdevctlCommands::Stop { uuid } => stop_command(&env, uuid),
+                force,
+            } => start_command(&env, uuid, parent, mdev_type, jsonfile, force),
+            MdevctlCommands::Stop { uuid, force } => stop_command(&env, uuid, force),
             MdevctlCommands::List(list) => list_command(
                 &env,
                 list.defined,
